# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/80_test_layers.ipynb (unless otherwise specified).

__all__ = ['ConvLayer', 'act_fn', 'Flatten', 'Noop', 'DownsampleBlock', 'BasicConvBlock', 'BasicBlock',
           'BottleConvBlock', 'Bottleneck', 'DownsampleLayer', 'ResBlock']

# Cell
import torch.nn as nn
import torch
from collections import OrderedDict

# Cell
act_fn = nn.ReLU(inplace=True)

class ConvLayer(nn.Sequential):
    """Basic conv layers block"""
    def __init__(self, ni, nf, ks=3, stride=1,
            act=True,  act_fn=act_fn,
            bn_layer=True, bn_1st=True, zero_bn=False,
            padding=None, bias=False, groups=1, **kwargs):

        self.act = act
        if padding==None: padding = ks//2
        layers = [('conv', nn.Conv2d(ni, nf, ks, stride=stride, padding=padding, bias=bias, groups=groups))]
        act_bn = [('act_fn', act_fn)] if act else []
        if bn_layer:
            bn = nn.BatchNorm2d(nf)
            nn.init.constant_(bn.weight, 0. if zero_bn else 1.)
            act_bn += [('bn', bn)]
        if bn_1st: act_bn.reverse()
        layers += act_bn
        super().__init__(OrderedDict(layers))

# Cell
class Flatten(nn.Module):
    '''flat x to vector'''
    def __init__(self):
        super().__init__()
    def forward(self, x): return x.view(x.size(0), -1)

# Cell
class Noop(nn.Module): # alternative name Merge
    '''Dummy module for vizualize skip conn'''
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x

# Cell
def DownsampleBlock(conv_layer, ni, nf, ks, stride, act=False,  **kwargs):
    '''Base downsample for res-like blocks'''
    return conv_layer(ni, nf, ks, stride, act, **kwargs)

class BasicConvBlock(nn.Sequential):
    '''Basic block of conv layers for resblock'''
    def __init__(self, conv_layer,ni,nh,nf,stride,zero_bn,**kwargs):
        super().__init__(OrderedDict([
            ('conv_0', conv_layer(ni, nh, 3, stride=stride, **kwargs)),
            ('conv_1', conv_layer(nh, nf, 3, zero_bn=zero_bn, act=False, **kwargs))
                                ]))


class BasicBlock(nn.Module):
    """Basic block (simplified) as in pytorch resnet"""
    def __init__(self, ni, nf,  expansion=1, stride=1,
                 bn_1st=True, zero_bn=False,
#                  groups=1, base_width=64, dilation=1, norm_layer=None
                conv_layer=ConvLayer, downsample_block=DownsampleBlock,
                 **kwargs):
        super().__init__()
        self.downsample = not ni==nf or stride==2
        self.conv = BasicConvBlock(conv_layer,ni,nf,nf,stride,zero_bn,**kwargs)
        if self.downsample:
            self.downsample = downsample_block(conv_layer, ni, nf, ks=1, stride=stride, act=False, **kwargs)
        self.merge = Noop()
        self.act_conn = act_fn

    def forward(self, x):
        identity = x
        out = self.conv(x)
        if self.downsample:
            identity = self.downsample(x)
        return self.act_conn(self.merge(out + identity))

# Cell

class BottleConvBlock(nn.Sequential):
    '''Basic block of conv layers for resblock'''
    def __init__(self, conv_layer,ni,nh,nf,stride,zero_bn,act=False,**kwargs):
        super().__init__(OrderedDict([
            ('conv_0', conv_layer(ni, nh, 1,                        **kwargs)),
            ('conv_1', conv_layer(nh, nh, 3, stride=stride,         **kwargs)),
            ('conv_2', conv_layer(nh, nf, 1, zero_bn=zero_bn, act=act, **kwargs))
                                ]))

class Bottleneck(nn.Module):
    '''Bottlneck block for resnet models'''
    def __init__(self, ni, nh, expansion=4, stride=1,
                 zero_bn=False,  # bn_1st=False,
                conv_layer=ConvLayer, act_fn=act_fn,
                 downsample_block=DownsampleBlock, **kwargs):
        super().__init__()
        self.downsample = not ni==nh or stride==2
        ni = ni*expansion
        nf = nh*expansion
        self.conv = BottleConvBlock(conv_layer,ni,nh,nf,stride,zero_bn,**kwargs)
        if self.downsample:
            self.downsample = downsample_block(conv_layer, ni, nf, ks=1, stride=stride, act=False, **kwargs)
        self.merge = Noop()
        self.act_conn = act_fn

    def forward(self, x):
        identity = x
        out = self.conv(x)
        if self.downsample:
            identity = self.downsample(x)
        return self.act_conn(self.merge(out + identity))

# Cell
class DownsampleLayer(nn.Sequential):
    """Downsample layer for Xresnet Resblock"""
    def __init__(self, ni, nf, stride=1,
                 conv_layer=ConvLayer, act_fn=act_fn,
                 pool=nn.AvgPool2d(2, ceil_mode=True), pool_1st=True,
                 **kwargs):
        layers  = [] if stride==1 else [('pool', pool)]
        layers += [] if ni==nf else [('idconv', conv_layer(ni, nf, 1, act=False, **kwargs))]
        if not pool_1st: layers.reverse()
        super().__init__(OrderedDict(layers))


class ResBlock(nn.Module):
    def __init__(self, ni, nh, expansion=1, stride=1, conv_layer=ConvLayer,
                 act_fn=act_fn, zero_bn=True, **kwargs):
        super().__init__()
        nf,ni = nh*expansion,ni*expansion
        self.convs = BasicConvBlock(conv_layer,ni,nh,nf,stride,zero_bn,**kwargs)\
                            if expansion==1 \
                            else BottleConvBlock(conv_layer,ni,nh,nf,stride,zero_bn,**kwargs)

        self.identity = DownsampleLayer(ni, nf, stride, **kwargs) if ni!=nf or stride==2 else Noop()
        self.merge = Noop() # us it to visualize in repr residual connection
        self.act_fn = act_fn

    def forward(self, x): return self.act_fn(self.merge(self.convs(x) + self.identity(x)))